import inspect
import json
from abc import ABC, abstractmethod
from typing import Any, Callable, Coroutine, List, Optional, Set, Type, TypeVar

import openai
from pydantic import BaseModel  # pylint: disable=no-name-in-module
from pydantic import create_model  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

from ..utils import Caser

F = TypeVar("F", bound="OpenAIFunction")
T = TypeVar("T")
P = ParamSpec("P")
c = Caser()

class FunctionCall(BaseModel):
    name: str
    data: Any

class OpenAIFunction(BaseModel, ABC):
    """Base class for OpenAI Functions"""
    class Metadata:
        subclasses: Set[Type[F]] = set()

    @classmethod
    def __init_subclass__(cls, **kwargs: Any):
        super().__init_subclass__(**kwargs)
        _schema = cls.schema()
        if cls.__doc__ is None:
            cls.__doc__ = ""
        cls.openaischema = {
            "name": cls.__name__,
            "description": cls.__doc__,
            "parameters": {
                "type": "object",
                "properties": {
                    k: v for k, v in _schema["properties"].items() if k != "self"
                },
                "required": _schema.get("required", []),
            },
        }
        cls.Metadata.subclasses.add(cls)

    @classmethod
    def from_function(cls, func: Callable[P, Coroutine[Any,Any,Any]]) -> Type["OpenAIFunction"]:
        kwargs = inspect.signature(func).parameters
        
        klass = create_model(
           c(func.__name__,"pascal"),
            __base__=cls,
            **{
                k: (v.annotation, ...) for k, v in kwargs.items() if k != "self"
            },
        )

        klass.__doc__ = func.__doc__
        
        
        async def run(self, **kwargs: Any) -> Any:
            return await func(self, **kwargs)
        
        setattr(klass, "run", run)
        klass.__init_subclass__()
        
        return klass
    
    async def __call__(self, **kwargs: Any) -> FunctionCall:
        response = await self.run(**kwargs)
        return FunctionCall(name=self.__class__.__name__, data=response)

    async def run(self, **kwargs: Any) -> Any:
        ...

async def parse_openai_function(
    response: dict[str, Any],
    functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses,
    **kwargs: Any,
) -> FunctionCall:
    choice = response["choices"][0]["message"]
    if "function_call" in choice:
        function_call_ = choice["function_call"]
        name = function_call_["name"]
        arguments = function_call_["arguments"]
        print(name, arguments)
        for i in functions:
            if i.__name__ == name:
                result = await i(**json.loads(arguments))(**kwargs)
                break
        else:
            raise ValueError(f"Function {name} not found")
        return result
    return FunctionCall(name="chat", data=choice["content"])


async def function_call(
    text: str,
    model: str = "gpt-3.5-turbo-16k-0613",
    functions: List[Type[F]] = OpenAIFunction.Metadata.subclasses,
    **kwargs: Any,
) -> FunctionCall:
    """Default function call"""
    messages = [
        {"role": "user", "content": text},
        {"role": "system", "content": "You are a function Orchestrator"},
    ]
    response = await openai.ChatCompletion.acreate( 
        model=model,
        messages=messages,
        functions=[i.openaischema for i in functions],
    ) 
    return await parse_openai_function(response, functions=functions, **kwargs)


async def chat_completion(text: str, context: Optional[str] = None):
    if context is not None:
        messages = [
            {"role": "user", "content": text},
            {"role": "system", "content": context},
        ]
    else:
        messages = [{"role": "user", "content": text}]
    response = await openai.ChatCompletion.acreate(
        model="gpt-3.5-turbo-16k-0613", messages=messages
    )
    return response["choices"][0]["message"]["content"]



from typing import List

import openai

from .vector import *


class Memory(VectorClient):
    """OpenAI Embeddings Client"""
    @property
    def builder(self):
        return QueryBuilder()
       
    async def encode(
        self, texts: List[str], namespace: str
    ) -> List[Embedding]:
        response = await openai.Embedding.acreate(  # type: ignore
            input=texts, model="text-embedding-ada-002"
        )
        response = await openai.Embedding.acreate(
            model="text-embedding-ada-002",
            input=texts,
        )
        data = response["data"]
        return [
            Embedding(
                values=embedding["embedding"],
                metadata={"namespace": namespace, "text": text},
            )
            for embedding, text in zip(data, texts)
        ]

    async def save(self, embeddings: List[Embedding]):
        response = await self.upsert(embeddings=embeddings)
        return response.upsertedCount

    async def search(
        self, text: str, namespace: str, top_k: int
    ) -> List[str]:
        
        embeddings = await self.encode(texts=[text], namespace=namespace)
        expr = (self.builder("namespace") == namespace) & (self.builder("text") != text)
        response = await self.query(
            expr=expr.query,
            vector=embeddings[0].values,
            topK=top_k,
        )
        return [match.metadata["text"] for match in response.matches]  # type: ignore



import re
from typing import Literal, TypeAlias

Case: TypeAlias = Literal["snake", "pascal", "camel", "constant", "human"]

class Caser:
	@staticmethod
	def snake_case(text:str):
		return re.sub(r'(?<!^)(?=[A-Z])', '_', text).lower()

	@staticmethod
	def pascal_case(text:str):
		return ''.join(word.capitalize() for word in text.split('_'))

	@staticmethod
	def camel_case(text:str):
		words = text.split('_')
		return words[0].lower() + ''.join(word.capitalize() for word in words[1:])

	@staticmethod
	def constant_case(text:str):
		return re.sub(r'(?<!^)(?=[A-Z])', '_', text).upper()

	@staticmethod
	def human_case(text:str):
		return ' '.join(word.capitalize() for word in text.split('_'))
	
	def __call__(self, text:str, case:Case):
		return getattr(self, f"{case}_case")(text)



import functools
from typing import Callable, Coroutine, Dict, List, Literal, TypeAlias, TypeVar

import openai  # pylint-disable=E0401
from aiofauna import *  # type: ignore
from jinja2 import Template
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module
from typing_extensions import ParamSpec

T = TypeVar("T")
P = ParamSpec("P")

Role: TypeAlias = Literal["assistant", "user", "system", "function"]
Model: TypeAlias = Literal["gpt-4-0613", "gpt-3.5-turbo-16k-0613"]
Size: TypeAlias = Literal["256x256", "512x512", "1024x1024"]
Format: TypeAlias = Literal["url", "base64"]


class Message(Document):
    """
    A message in a chat

    Args:
        role (Role): The role of the message (assistant, user, system, function)
        content (str): The content of the message
    """
    role: Role = Field(..., description="The role of the message")
    content: str = Field(..., description="The content of the message")


class ChatCompletionRequest(Document):
    """

    Chat completion request

    Args:

        model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)
        messages (List[Message]): The messages to use for the completion
        temperature (float): The temperature of the completion
        max_tokens (int): The maximum number of tokens to generate
        stream (bool): Whether to stream the completion or not
    """
    model: Model = Field(..., description="The model to use for the completion")
    messages: List[Message] = Field(
        ..., description="The messages to use for the completion"
    )
    temperature: float = Field(
        default=0.5, description="The temperature of the completion"
    )
    max_tokens: int = Field(
        1024, description="The maximum number of tokens to generate"
    )
    stream: bool = Field(False, description="Whether to stream the completion or not")


class ChatCompletionUssage(Document):
    """

    Chat completion usage

    Args:

        prompt_tokens (int): The number of tokens in the prompt

        completion_tokens (int): The number of tokens in the completion

        total_tokens (int): The total number of tokens

    """
        
    prompt_tokens: int = Field(..., description="The number of tokens in the prompt")
    completion_tokens: int = Field(
        ..., description="The number of tokens in the completion"
    )
    total_tokens: int = Field(..., description="The total number of tokens")


class ChatCompletionChoice(Document):
    """

    Chat completion choice

    Args:

        index (int): The index of the choice

        message (Message): The message of the choice

        finish_reason (str): The reason the choice was finished

    """ 
    
    index: int = Field(..., description="The index of the choice")
    message: Message = Field(..., description="The message of the choice")
    finish_reason: str = Field(..., description="The reason the choice was finished")


class ChatCompletionResponse(Document):
    """

    Chat completion response

    Args:

        id (str): The id of the completion

        object (str): The object of the completion

        created (int): The creation time of the completion

        model (Model): The model used for the completion

        choices (List[ChatCompletionChoice]): The choices of the completion

        usage (ChatCompletionUssage): The usage of the completion

        stream (bool): Whether the completion was streamed or not

    """

    id: str = Field(..., description="The id of the completion")
    object: str = Field(..., description="The object of the completion")
    created: int = Field(..., description="The creation time of the completion")
    model: Model = Field(..., description="The model used for the completion")
    choices: List[ChatCompletionChoice] = Field(
        ..., description="The choices of the completion"
    )
    usage: ChatCompletionUssage = Field(..., description="The usage of the completion")
    stream: bool = Field(..., description="Whether the completion was streamed or not")


class VectorResponse(Document):
    """

    Vector response

    Args:

        id (str): The id of the vector

        object (str): The object of the vector

        created (int): The creation time of the vector

        model (Model): The model used for the vector

        data (List[float]): The data of the vector

    """  
    
    text: str = Field(..., description="The text of the completion")
    score: float = Field(..., description="The score of the completion")


class CreateImageResponse(Document):
    """

    Create image response

    Args:

        id (str): The id of the image

        object (str): The object of the image

        created (float): The creation time of the image

        data (List[Dict[Format, str]]): The data of the image

    """
    
    created: float = Field(...)
    data: List[Dict[Format, str]] = Field(...)


class CreateImageRequest(Document):
    """

    Create image request

    Args:

        prompt (str): The prompt of the image

        n (int): The number of images to generate

        size (Size): The size of the image

        response_format (Format): The format of the response

    """   
    
    prompt: str = Field(...)
    n: int = Field(default=1)
    size: Size = Field(default="1024x1024")
    response_format: Format = Field(default="url")


class LanguageModel(BaseModel):
    """

    ChatGPT: A GPT-3 chatbot

    This class is a wrapper around the OpenAI Chat API

    Args:

        model (Model): The model to use for the completion (gpt-4-0613, gpt-3.5-turbo-16k-0613)

        temperature (float): The temperature of the completion

        max_tokens (int): The maximum number of tokens to generate

    """  
    
    
    model: Model = Field(default="gpt-4-0613")
    temperature: float = Field(default=0.2)
    max_tokens: int = Field(default=1024)

    async def chat(
        self, messages: List[Message]
    ) -> str:
        """

        Chat with the bot

        Args:

            messages (List[Message]): The messages to use for the completion

        Returns:

            str: The response of the bot

        """
       
        response = await openai.ChatCompletion.acreate(  # type: ignore
            model=self.model,
            messages=[message.dict() for message in messages],
            temperature=self.temperature,
            max_tokens=self.max_tokens,
            stream=False
        )
        
        return response.choices[0].message.content  # type: ignore

    def __call__(self, func: Callable[P,Coroutine[T,Any,Any]]) -> Callable[P,Coroutine[T,Any,Any]]:
        """

        Call the bot with a function that features a jinja2 docstring as a template for the system message, and the user message as the first argument, prompt engineering is done automatically

        Args:

            func (Callable[P,Coroutine[T,Any,Any]]): The function to call

        Returns:

            Callable[P,Coroutine[T,Any,Any]]: The wrapped function

        """
        if func.__doc__ is None:
            raise ValueError("Function must have a docstring")
        _template = func.__doc__	
        @handle_errors
        @functools.wraps(func)
        async def wrapper(*args:P.args, **kwargs:P.kwargs) -> T:
            template = Template(_template)
            ctx = template.render(**kwargs)
            kwargs["context"] = ctx
            return await func(*args, **kwargs)
        return wrapper


from pydantic import BaseConfig

from .memory import *
from .tools import *


class Agent(ABC):
	
	
	llm:LanguageModel
	memory:Memory
	namespace:str
	n:int
	

	@abstractmethod
	async def __call__(self, text:str)->str:
		"""
		Chain a text
		"""
		...

class Swagchain(Agent):

	def __init__(self, llm:LanguageModel=LanguageModel(), memory:Memory=Memory(), namespace:str="main", n:int=5):
		self.llm = llm
		self.memory = memory
		self.namespace = namespace
		self.n = n

	async def __call__(self, text:str):
		"""
		Chain a text
		"""
		matches = await self.memory.search(text=text, namespace=self.namespace, top_k=self.n)
		if matches:
			context = "Suggestions:\n"+"\n".join(matches)
			return await self.llm.chat(messages=[Message(role="user",content=text), Message(role="system",content=context)])
		else:
			return await self.llm.chat(messages=[Message(role="user",content=text)])
	



import asyncio
import os
from typing import AsyncGenerator, List

from aiofauna import handle_errors, process_time, setup_logging  # type: ignore
from aiohttp import ClientSession, TCPConnector
from bs4 import BeautifulSoup  # pylint: disable=import-error

from ..llm import *
from ..memory import *

logger = setup_logging(__name__)

HEADERS = {
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/91.0.4472.114 Safari/537.36"
}
BAD_EXT = (
    "png",
    "jpg",
    "jpeg",
    "gif",
    "pdf",
    "doc",
    "docx",
    "ppt",
    "pptx",
    "xls",
    "xlsx",
    "zip",
    "rar",
    "gz",
    "7z",
    "exe",
    "mp3",
    "mp4",
    "avi",
    "mkv",
    "mov",
    "wmv",
    "flv",
    "swf",
)

connector = TCPConnector(limit=1000)

logger = setup_logging(__name__)
embeddings = Memory(
    api_endpoint=os.environ.get("PINECONE_API_URL"), api_key=os.environ.get("PINECONE_API_KEY")  # type: ignore
)


@process_time
@handle_errors
async def sitemap(url: str, session: ClientSession) -> List[str]:
    urls: List[str] = []
    if not url.endswith("xml"):
        url = f"{url.rstrip('/')}/sitemap.xml"
    async with session.get(url) as response:
        text = await response.text()
        soup = BeautifulSoup(text, features="xml")
        for loc in soup.findAll("loc"):
            if loc.text.endswith(BAD_EXT):
                continue
            urls.append(loc.text)
            logger.info("Found url: %s", loc.text)
        for nested_sitemap in soup.findAll("sitemap"):
            urls.extend(await sitemap(nested_sitemap.loc.text, session))
    return urls


@process_time
@handle_errors
async def fetch_website(url: str, session: ClientSession, max_size: int = 40960) -> str:
    async with session.get(url) as response:
        html = await response.text()
        soup = BeautifulSoup(html, features="lxml")
        tag = soup.find("div", class_="content")
        assert tag is not None
        data = tag.get_text(strip=True)
        logger.info("Fetched %s", url)
        logger.info("Length: %s", len(data))
        return data[:max_size]


async def sitemap_pipeline(
    url: str,
    namespace: str,
    session: ClientSession,
    chunk_size: int = 64,
):
    urls: List[str] = await sitemap(url, session)
    length: int = len(urls)
    inserted: int = 0
    while urls:
        try:
            chunk = urls[:chunk_size]
            urls = urls[chunk_size:]
            contents = await asyncio.gather(
                *[fetch_website(url, session) for url in chunk]
            )
            inserted += await embeddings.save(
                [
                    Embedding(
                        values=content, metadata={"namespace": namespace, "text": text}
                    )
                    for content, text in zip(contents, chunk)
                ]
            )
            progress = inserted / length
            logger.info("Progress: %s", progress)
            yield progress
            if progress >= 1:
                yield 1
                logger.info(f"Succesfully inserted {inserted} documents")
                break
        except Exception as e:
            logger.error(e)
            continue


@process_time
@handle_errors
async def find_all_children(url: str, session: ClientSession) -> List[str]:
    async with session.get(url + "/index.html") as response:
        html: str = await response.text()
        soup = BeautifulSoup(html, features="lxml")
        results = soup.find_all("a", href=True)
        logger.info("Found %s results", len(results))
        responses: List[str] = []
        for result in results:
            if result["href"].startswith("http"):
                responses.append(result["href"])
            else:
                responses.append(f"{url.rstrip('/')}/{result['href'].lstrip('/')}")
        return responses


@process_time
@handle_errors
async def fetch_all_children_recursive(url: str, session: ClientSession) -> List[str]:
    urls = await find_all_children(url, session)
    for url in urls:
        urls.extend(await fetch_all_children_recursive(url, session))
    return urls


async def fetch_pipeline(
    url: str,
    namespace: str,
    session: ClientSession,
    chunk_size: int = 64,
):
    urls = await find_all_children(url, session)
    logger.info("Found %s urls", len(urls))
    logger.info("Fetching all children")
    length = len(urls)
    inserted = 0
    while urls:
        chunk = urls[:chunk_size]
        urls = urls[chunk_size:]
        try:
            contents = await asyncio.gather(
                *[fetch_website(url, session) for url in chunk]
            )
            inserted += await embeddings.save(
                [
                    Embedding(
                        values=content, metadata={"namespace": namespace, "text": text}
                    )
                    for content, text in zip(contents, chunk)
                ]
            )
            progress = inserted / length
            logger.info("Progress: %s", progress)
            yield progress
            if progress >= 1:
                logger.info(f"Succesfully inserted {inserted} documents")
                break
        except Exception as e:
            logger.error(e)
            continue


async def website_loader(url: str, namespace: str) -> AsyncGenerator[float, None]:
    """Reads a website and returns a list of strings"""
    async with ClientSession(connector=connector, headers=HEADERS) as session:
        exts = ["sitemap.xml", "index.html", ""]
        for e in exts:
            response = await session.get(url + "/" + e)
            if response.status == 200:
                if e == "sitemap.xml":
                    async for progress in sitemap_pipeline(url, namespace, session):
                        yield progress
                else:
                    async for progress in fetch_pipeline(url, namespace, session):
                        yield progress
                break
        else:
            logger.error("Could not find a sitemap.xml or index.html file")
            raise Exception("Could not find a sitemap.xml or index.html file")



"""Swagchain `vector` module: A mini sdk of Pinecone's vector database API."""
from __future__ import annotations

import os
from typing import Any, Dict, List, Literal, TypeAlias, Union
from uuid import uuid4

from aiofauna import LazyProxy  # type: ignore
from aiohttp import ClientSession
from pydantic import BaseModel, Field  # pylint: disable=no-name-in-module

Vector: TypeAlias = List[Union[int, float]]
Value: TypeAlias = Union[int, float, str, bool, List[str]]
QueryJoin: TypeAlias = Literal["$and", "$or"]
QueryWhere: TypeAlias = Literal[
    "$eq", "$ne", "$gt", "$gte", "$lt", "$lte", "$in", "$nin"
]
QueryKey: TypeAlias = Union[str, QueryWhere, QueryJoin]
QueryValue: TypeAlias = Union[Value, List[Value], "Query", List["Query"]]
Query: TypeAlias = Dict[QueryKey, QueryValue]
MetaData: TypeAlias = Dict[str, Value]
# Please write Sphinx documentation for this m

class QueryBuilder(object):
    
    def __init__(self, field: str = None, query: Query = None):  # type: ignore
        """
        Initializes a new QueryBuilder instance.

        Args:
            field (str, optional): The field to query. Defaults to None.
            query (Query, optional): The query dictionary. Defaults to an empty dictionary.
        """
        self.field = field
        self.query = query if query else {}

    def __repr__(self) -> str:
        """Returns the string representation of the query."""
        return f"{self.query}"

    def __call__(self, field_name: str) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with a given field name.

        Args:
            field_name (str): The field name to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance.
        """
        return QueryBuilder(field_name)

    def __and__(self, other: QueryBuilder) -> QueryBuilder:
        """
        Combines two queries using the $and operator.

        Args:
            other (QueryBuilder): Another QueryBuilder instance.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        """
        return QueryBuilder(query={"$and": [self.query, other.query]})

    def __or__(self, other: QueryBuilder) -> QueryBuilder:
        """
        Combines two queries using the $or operator.

        Args:
            other (QueryBuilder): Another QueryBuilder instance.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        """
        return QueryBuilder(query={"$or": [self.query, other.query]})

    def __eq__(self, value: Value) -> QueryBuilder:  # type: ignore
        """
        Creates a new QueryBuilder instance with the $eq operator.

        Args:
            value (Value): The value to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        
        """
        return QueryBuilder(query={self.field: {"$eq": value}})

    def __ne__(self, value: Value) -> QueryBuilder:  # type: ignore
        """
        Creates a new QueryBuilder instance with the $ne operator.

        Args:
            value (Value): The value to query.

        Returns:
            QueryBuilder: A new QueryBuilder instance with the combined query.
        
        """
        return QueryBuilder(query={self.field: {"$ne": value}})

    def __lt__(self, value: Value) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with the $lt operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.
        
        
        """
        
        return QueryBuilder(query={self.field: {"$lt": value}})

    def __le__(self, value: Value) -> QueryBuilder:
        """
        Creates a new QueryBuilder instance with the $lte operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$lte": value}})

    def __gt__(self, value: Value) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $gt operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """
        
        return QueryBuilder(query={self.field: {"$gt": value}})

    def __ge__(self, value: Value) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $gte operator.

        Args:

            value (Value): The value to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.
        
        """
        
        
        return QueryBuilder(query={self.field: {"$gte": value}})

    def in_(self, values: List[Value]) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $in operator.

        Args:

            values (List[Value]): The values to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """
            
        return QueryBuilder(query={self.field: {"$in": values}})

    def nin_(self, values: List[Value]) -> QueryBuilder:
        """

        Creates a new QueryBuilder instance with the $nin operator.

        Args:

            values (List[Value]): The values to query.

        Returns:

            QueryBuilder: A new QueryBuilder instance with the combined query.

        """

        return QueryBuilder(query={self.field: {"$nin": values}})

class UpsertRequest(BaseModel):
    """
    A request to upsert a vector.

    Args:
        id (str, optional): The ID of the vector. Defaults to a random UUID.
        values (Vector): The vector values.
        metadata (MetaData): The vector metadata.
    """
    id: str = Field(default_factory=lambda: str(uuid4()))
    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class Embedding(BaseModel):
    """

    A vector embedding.

    Args:

        values (Vector): The vector values.

        metadata (MetaData): The vector metadata.

    """


    values: Vector = Field(...)
    metadata: MetaData = Field(...)


class QueryRequest(BaseModel):
    """
    A request to query vectors.

    Args:

        topK (int, optional): The number of results to return. Defaults to 10.

        filter (Query): The query filter.

        includeMetadata (bool, optional): Whether to include metadata in the response. Defaults to True.

        vector (Vector): The vector to query.

    """
    topK: int = Field(default=10)
    filter: Dict[str, Any] = Field(...)
    includeMetadata: bool = Field(default=True)
    vector: Vector = Field(...)


class QueryMatch(BaseModel):
    """

    A query match.

    Args:

        id (str): The ID of the vector.

        score (float): The match score.

        metadata (MetaData): The vector metadata.

    """  
    
    id: str = Field(...)
    score: float = Field(...)
    metadata: MetaData = Field(...)


class QueryResponse(BaseModel):
    """

    A response to a query request.

    Args:

        matches (List[QueryMatch]): The query matches.

    """   
    matches: List[QueryMatch] = Field(...)


class UpsertResponse(BaseModel):
    """

    A response to an upsert request.

    Args:

        upsertedCount (int): The number of vectors upserted.

    """ 
    
    upsertedCount: int = Field(...)


class VectorClient(LazyProxy[ClientSession]):
    """

    A client for the vector database.

    Args:

        api_key (str): The API key.

        api_endpoint (str): The API endpoint.

    """
    def __init__(self, api_key: str=os.environ.get("PINECONE_API_KEY"), api_endpoint: str=os.environ.get("PINECONE_API_URL")) -> None:  # type: ignore
        self.api_key = api_key
        self.api_endpoint = api_endpoint
        super().__init__()

    def __load__(self) -> ClientSession:
        """

        Loads the client session.

        Returns:

            ClientSession: The client session.

        """
               
        return ClientSession(
            headers={"api-key": self.api_key},
            base_url=self.api_endpoint,
        )

    async def upsert(self, embeddings: List[Embedding]) -> UpsertResponse:
        """

        Upserts a list of vector embeddings.

        Args:

            embeddings (List[Embedding]): The vector embeddings.

        Returns:

            UpsertResponse: The upsert response.

        """        
        async with self.__load__() as session:
            values: List[Vector] = []
            metadata: List[MetaData] = []
            for embedding in embeddings:
                values.append(embedding.values)
                metadata.append(embedding.metadata)

            async with session.post(
                "/vectors/upsert",
                json={
                    "vectors": [
                        UpsertRequest(values=values, metadata=metadata).dict()
                        for values, metadata in zip(values, metadata)
                    ]
                },
            ) as response:
                return UpsertResponse(**await response.json())

    async def query(
        self, expr: Query, vector: Vector, topK: int, includeMetadata: bool = True
    ) -> QueryResponse:
        """

        Queries the vector database.

        Args:

            expr (Query): The query expression.

            vector (Vector): The vector to query.

            topK (int): The number of results to return.

            includeMetadata (bool, optional): Whether to include metadata in the response. Defaults to True.

        Returns:

            QueryResponse: The query response.

        """

        async with self.__load__() as session:
            payload = QueryRequest(
                topK=topK,
                filter=expr,  # type: ignore
                vector=vector,
                includeMetadata=includeMetadata,
            ).dict()
            async with session.post(
                "/query",
                json=payload,
            ) as response:
                return QueryResponse(**await response.json())